//  Copyright (c) 2006, Wade Tregaskis.  All rights reserved.
//  Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
//    * Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
//    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
//    * Neither the name of Wade Tregaskis nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


#ifndef SORTALGORITHMS_H
#define SORTALGORITHMS_H

#include <vector>
#include "PerformanceMetrics.h"
using namespace std;


// AVL tree sort
//
// Uses the AVL tree class to sort everything into a balanced ordered tree,
// and then flatten the result back into an ordered vector.  The overhead
// of converting from and back to a vector is significant, but largely linear.

template<class V, class A> void performAVLTreeSort(vector<V*, A> &vec) {
    AVLTree<V*> tree;
    typename vector<V*, A>::iterator iterator, endOfVector = vec.end();
    register unsigned int originalSize;

    for (iterator = vec.begin(); iterator < endOfVector; ++iterator) {
        tree.insert(*iterator);
    }

    originalSize = vec.size();
    vec.clear();
    vec.reserve(originalSize);

    treeInOrderToVector(tree.getRoot(), vec);
}


// Optimized Radix sort
//
// The secret with this particular application is that the list as received
// is already sorted.  All we want to do is resort it with the postcode as
// the primary key field, but otherwise in the same order as it's already in.
//
// So, using a stable variety of the radix sort (as most are) we need only
// sort via the postcode, and then concatenate these back into a final, sorted
// list.  So we don't really do a true radix sort - more like a hash table, or
// even more accurately just a simple array index.

template<class V, class A> void performOptimizedRadixSort(vector<V*, A> &vec) {
    vector<V*, A> array[10000];
    typename vector<V*, A>::const_iterator constEnd, constIterator;
    typename vector<V*, A>::iterator mutableEnd, mutableIterator;
    register V *temp;
    register unsigned int size;
    register int i;

    constEnd = vec.end();

    for (constIterator = vec.begin(); constIterator < constEnd; ++constIterator) {
        temp = *constIterator;
        array[temp->postcode()].push_back(temp);
    }

    size = vec.size();
    insertCount += size;
    removeCount += size;
    vec.erase(vec.begin(), vec.end());
    vec.reserve(size);

    for (i = 0; i != 10000; ++i) {
        if (!array[i].empty()) {
            mutableEnd = array[i].end();

            for (mutableIterator = array[i].begin(); mutableIterator < mutableEnd; ++mutableIterator) {
                vec.push_back(*mutableIterator);
            }

            size = array[i].size();
            insertCount += size;
            removeCount += size;
            array[i].erase(array[i].begin(), array[i].end());
        }
    }
}

// Merge sort
//
// 1) Divide the array in half and merge sort both sides
// 2) Combine the results from both sides in sorted order

// Note that while I could use the STL merge algorithm to do this,
// I wasn't sure if it was allowed.  I felt it best to write my own,
// anyway, as it's all good experience.

template<class V, class A, typename I> vector<V*, A>* merge(vector<V*, A> &original, I startA, I endA, I startB, I endB) {
    register vector<V*, A> *result = new vector<V*, A>;
    I currentA(startA), currentB(startB);

    result->reserve((endA - startA) + (endB - startB));

    while ((currentA < endA) && (currentB < endB)) {
        if (*(*currentA) < *(*currentB)) {
            result->push_back(*currentA);
            ++currentA;
        } else {
            result->push_back(*currentB);
            ++currentB;
        }
    }

    while (currentA < endA) {
        result->push_back(*currentA);
        ++currentA;
    }

    while (currentB < endB) {
        result->push_back(*currentB);
        ++currentB;
    }

    return result;
}

template<class V, class A, typename I> vector<V*, A>* mergeSort(vector<V*, A> &original, I start, I end) {
    I middle;
    vector<V*, A> *resultA, *resultB, *result;

    if (start < (end - 1)) {
        middle = start + ((end - start) / 2);

        resultA = mergeSort(original, start, middle);
        resultB = mergeSort(original, middle, end);

        result = merge(original, resultA->begin(), resultA->end(), resultB->begin(), resultB->end());

        delete resultA;
        delete resultB;

        return result;
    } else {
        return new vector<V*, A>(start, end);
    }
}

template<class V, class A> void performMergeSort(vector<V*, A> &vec) {
    vector<V*, A> *result;

    result = mergeSort(vec, vec.begin(), vec.end());

    vec.clear();
    vec.reserve(result->size());
    vec = *result;

    delete result;
}


// Quick sort
//
// 1) Parition the array and create a pivot point
// 2) Sort both sides of the pivot using Quick sort
//
// Clearly this is a highly recursive function
//
// It was also rather tricky to implement - iterators
// are not fun to play with... so many off by one
// errors.

template<typename I> I quickPartition(I start, I end) {
    I left, right;

    left = start + 1;
    right = end - 1;

    while (left <= right) {
        while ((left < end) && (*(*left) <= *(*start))) {
            ++left;
        }

        while ((right >= start) && (*(*right) > *(*start))) {
            --right;
        }

        if (left < right) {
            swap(*left, *right);
        } else {
            ++left;
        }
    }

    if (right != start) {
        swap(*right, *start);
    }

    return right;
}

template<typename I> void quickSort(I start, I end) {
    I middle;

    if (start < end) {
        middle = quickPartition(start, end);

        quickSort(start, middle);
        quickSort(middle + 1, end);
    }
}

template<class V, class A> void performQuickSort(vector<V*, A> &vec) {
    quickSort(vec.begin(), vec.end());
}


// Bubble sort
//
// Start with the first item, and compare it with the next.  If they
// are in the wrong order, swap them.  Repeat for the next item.  Each
// pass must be from the first item to the last, then the second last,
// then the third last, etc.

template<class V, class A> void performBubbleSort(vector<V*, A> &vec) {
    register int rearOffset = 1;
    typename vector<V*, A>::iterator lastItem, iterator;

    while ((lastItem = vec.end() - rearOffset++) != vec.begin()) {
        for (iterator = vec.begin(); iterator < lastItem; ++iterator) {
            if (*(*iterator) > *(*(iterator + 1))) {
                swap(*iterator, *(iterator + 1));
            }
        }
    }
}


// Selection Sort
//
// Start with the first item, and select it.  Go through to the last item,
// then the second last, etc, on each iteration.  Compare the selected
// item with the current one.  If the current one is bigger, make it the
// selected item.  At the end, swap the selected item with the last item.

template<class V, class A> void performSelectionSort(vector<V*, A> &vec) {
    register int rearOffset = 1;
    typename vector<V*, A>::iterator lastItem, iterator, selectedItem;

    while ((lastItem = vec.end() - rearOffset++) != vec.begin()) {
        for (iterator = (selectedItem = vec.begin()) + 1; iterator < lastItem + 1; ++iterator) {
            if (*(*selectedItem) < *(*iterator)) {
                selectedItem = iterator;
            }
        }

        if (selectedItem != lastItem) {
            swap(*selectedItem, *lastItem);
        }
    }
}


// Insertion Sort
//
// Start with the second item.  Go through to the last item.
// For each item, proceed backwards through the array until you
// find an item that is smaller than it.  Insert it after this
// item.

template<class V, class A> void performInsertionSort(vector<V*, A> &vec) {
    register int masterOffset, sizeOfVec;
    typename vector<V*, A>::iterator iterator;
    register V *selectedItem;
    register bool didInsert;

    sizeOfVec = vec.size();

    for (masterOffset = 1; masterOffset < sizeOfVec; ++masterOffset) {
        iterator = vec.begin() + masterOffset;
        selectedItem = *iterator;

        if (*selectedItem < *(*(iterator - 1))) {
            remove(vec, iterator);

            didInsert = false;

            for (iterator = vec.begin() + (masterOffset - 2); !didInsert && (iterator >= vec.begin()); --iterator) {
                if (*(*iterator) < *selectedItem) {
                    insert(vec, iterator + 1, selectedItem);

                    didInsert = true;
                }
            }

            if (!didInsert) {
                insert(vec, vec.begin(), selectedItem);
            }
        }
    }
}

#endif
